# import torch
# from Fireworks import dataset as ds
#
# def test_get_next_n(): pass
#
# def test_FireworksDataset(): pass
#
# def test_get_chunk(): pass
#
# def test_to_tensor(): pass
#
# def test_iterate(): pass
