import torch

class FireworksDataLoader:
    """
    Class for generating batches from datasets. This class performs tasks such as
    orchestrating batches accross multiple datasets, applying transforms to those batches,
    and over/unersampling
    """
    pass
